import asyncio
from typing import List, Optional
from langchain_core.embeddings import Embeddings
from langchain_core.pydantic_v1 import BaseModel

from ascenderai.ascender_ai import AscenderAI

class AscenderAIEmbeddingsL(BaseModel, Embeddings):
    model_name: str = "text-embedder-minilm-l6v2"
    output_type: str = "sentence_embedding"
    verbose: bool = False

    base_url: Optional[str] = None

    def embed_documents(self, texts: list[str]) -> list[list[float]]:
        client = AscenderAI(self.model_name, verbose=self.verbose)

        if self.base_url:
            client.base_url = self.base_url
        # NOTE: This is a blocking call.
        _response = asyncio.run(client.embeddings.create(texts, output_type=self.output_type))

        return _response.data.tolist()[0]
    
    def embed_query(self, text: str) -> List[float]:
        client = AscenderAI(self.model_name, verbose=self.verbose)
        
        if self.base_url:
            client.base_url = self.base_url
        # NOTE: This is a blocking call.
        _response = asyncio.run(client.embeddings.create(text, output_type=self.output_type))

        return _response.data.tolist()[0]
    
    async def aembed_documents(self, texts: list[str]) -> list[list[float]]:
        client = AscenderAI(self.model_name, verbose=self.verbose)
        if self.base_url:
            client.base_url = self.base_url
        # Response in formated instance
        _response = await client.embeddings.create(texts, output_type=self.output_type)

        return _response.data.tolist()

    async def aembed_query(self, text: str) -> List[float]:
        client = AscenderAI(self.model_name, verbose=self.verbose)

        if self.base_url:
            client.base_url = self.base_url
        # Response in formated instance
        _response = await client.embeddings.create(text, output_type=self.output_type)

        return _response.data.tolist()


from .ascender_ai import AscenderAI
from .langchain_llm import AscenderAIL
from .langchain_embedding import AscenderAIEmbeddingsL

__version__ = "0.1.0"

__all__ = [AscenderAI, AscenderAIL, AscenderAIEmbeddingsL, __version__]


from typing import Literal, Optional
from numpy import ndarray
from pydantic import BaseModel


class CompletionChoice(BaseModel):
    text: str
    index: int = 0
    logprobs: Optional[dict]
    finish_reason: Optional[Literal["length", "stop"]] = None


class CompletionUsage(BaseModel):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class CompletionResponse(BaseModel):
    id: str
    model: str = "mistral-7b-instruct"
    object: str = "text_completion"
    created: int
    choices: list[CompletionChoice]
    usage: Optional[CompletionUsage] = None


class EmbeddingUsage(BaseModel):
    prompt_tokens: int
    total_tokens: int


class EmbeddingResponse(BaseModel):
    id: str
    model: str = "text-embedder-minilm-l6v2"
    object: Literal["embeddings.sentence", "embeddings.token"]
    data: ndarray | bytes
    usage: EmbeddingUsage

    class Config:
        arbitrary_types_allowed = True
        json_encoders = {
            ndarray: lambda v: v.tolist()
        }


import json
from aiohttp import ClientSession
from typing import AsyncIterator, Literal, Optional
from rich import print as rprint
from rich.panel import Panel
from numpy import array

import os
from ascenderai.handler import AscenderExceptionManager

from ascenderai.types import CompletionChoice, CompletionResponse, CompletionUsage, EmbeddingResponse, EmbeddingUsage


class AscenderAI:

    def __init__(self, model: str = "mistral-7b-instruct", *, max_tokens: int = 200,
                 streaming: bool = False, verbose: bool = False) -> None:
        self.completions = AscenderAIChat(
            model=model,
            max_tokens=max_tokens,
            streaming=streaming,
            verbose=verbose
        )
        self.embeddings = AscenderAIEmbeddings(
            model=model,
            verbose=verbose
        )
    @property
    def base_url(self):
        return self.completions.base_url
    
    @base_url.setter
    def base_url(self, value: str):
        self.completions.base_url = value
        self.embeddings.base_url = value

    def ping(self):
        """
        This method made for testing connection with AI Cluster and health of AI Cluster.
        """
        # TODO: Implement the ping-pong method
        return "PONG!"
    
    def __str__(self) -> str:
        return self.completions.model
    
    def __repr__(self) -> str:
        return f"AscenderAI(model={self.completions.model!r}, max_tokens={self.completions.max_tokens}, streaming={self.completions.streaming}, verbose={self.completions.verbose})"


class AscenderAIChat:
    base_url: str = os.getenv("ASCENDER_AI_BASE_URL", "https://ai.ascender.space")
    def __init__(self, model: str = "mistral-7b-instruct", *,
                 max_tokens: int = 30, streaming: bool = False,
                 verbose: bool = False) -> None:
        self.model = model
        self.max_tokens = max_tokens
        self.streaming = streaming
        self.verbose = verbose
        self._exception_manager = AscenderExceptionManager(model)
    
    async def create(self, prompt: str, *, temperature: float = 0.9, 
                             stop_seqs: Optional[list[str]] = None, seed: Optional[int] = None) -> CompletionResponse | AsyncIterator[CompletionResponse]:
        """
        ## Create Completion

        `AscenderAIChat.create()` or `AscenderAI.completions.create()` is used to create completions for a given prompt.
        It creates a completion for the given prompt using the specified LLM model.

        Args:
            prompt (str): Prompt for which the completion will be created.
            temperature (float, optional): Temperature of inference during sampling, from 0 to 1. Less temperature - more creative. Defaults to 0.9.
            stop_seqs (Optional[list[str]], optional): Tokens where inference response generation should be stopped. Defaults to None.
            seed (Optional[int], optional): Seed of inference. Defaults to None.
        """
        # Verbose displaying of the prompt
        if self.verbose:
            panel = Panel(f"[bold green]{prompt.replace('[', '\[').replace("]", "\]")}[/bold green]", title="Get completion request")
            rprint(panel)
        
        if self.streaming:
            return self.stream(prompt, temperature=temperature, stop=stop_seqs, seed=seed)
        else:
            async with ClientSession(base_url=self.base_url) as client:
                async with client.post("/completions/completions", json={
                    "prompt": prompt,
                    "model": self.model,
                    "max_tokens": self.max_tokens,
                    "temperature": temperature,
                    "stop_seqs": stop_seqs,
                    "seed": seed,
                }) as resp:
                    # For HTTP error handling
                    self._exception_manager.listen_for_errors("completions.create", resp)

                    # NOTE: If http error will be raised, then exception manager will stop execution of code from this line
                    response = await resp.json()
                    choices = response["choices"]
                    usage = response["usage"]

                    if self.verbose:
                        response_panel = Panel(f"[bold yellow]{response['choices'][0]['text']}[/bold yellow]", title="AI Response")
                        rprint(response_panel)
                    
                    del response["choices"]
                    del response["usage"]

                    response = CompletionResponse(
                        **response, choices=[CompletionChoice(**choice) for choice in choices],
                        usage=CompletionUsage(**usage)
                    )
                    return response
    
    async def stream(self, prompt: str, *, temperature: float = 0.9,
                                       stop: Optional[list[str]] = None, seed: Optional[int] = None):
        """
        ## Stream Completions

        `AscenderAIChat.stream()` or `AscenderAI.completions.stream()` is similar to `AscenderAIChat.stream()` or `AscenderAI.completions.stream()`.
        When previous one waits until completion would be generated completely and then returns, this one streams by returning iterator with token chunks.
        """
        
        async with ClientSession(base_url=self.base_url) as client:
            async with client.post("/completions/completions/stream", json={
                "prompt": prompt,
                "model": self.model,
                "max_tokens": self.max_tokens,
                "temperature": temperature,
                "stop_seqs": stop,
                "seed": seed,
            }) as resp:
                # For HTTP error handling
                self._exception_manager.listen_for_errors("completions.stream", resp)

                # NOTE: If http error will be raised, then exception manager will stop execution of code from this line
                async for line in resp.content.iter_any():
                    try:
                        readed_line = line.decode()

                        readed_line.split("} {")
                        json_decoded = json.loads(readed_line)
                        
                        choices = json_decoded["choices"]
                        usage = json_decoded["usage"]
                        
                        del json_decoded["choices"]
                        del json_decoded["usage"]
                        
                        response = CompletionResponse(
                            **json_decoded, choices=[CompletionChoice(**choice) for choice in choices],
                            usage=CompletionUsage(**usage)
                        )
                        yield response

                    except:
                        pass
                
    

class AscenderAIEmbeddings:

    base_url: str = os.getenv("ASCENDER_AI_BASE_URL", "https://ai.ascender.space")
    def __init__(self, model: str = "text-embedder-minilm-l6v2", *,
                 verbose: bool = False) -> None:
        self.model = model
        self.verbose = verbose
        self._exception_manager = AscenderExceptionManager(model)
    
    async def create(self, prompt: str | list[str], 
                     encoding_format: Literal["base64", "float"] = "float",
                     output_type: Optional[Literal["sentence_embedding", "token_embeddings"]] = "sentence_embedding") -> EmbeddingResponse:
        """
        ## Create Embedding

        `AscenderAIEmbeddings.create()` or `AscenderAI.embeddings.create()` is used to create embeddings for a given sentence or list of sentences.
        It creates vectoral representration of the sentence(s) using the specified sentence embedding model.

        Can be used for:
        - Search (where results are ranked by relevance to a query string)
        - Clustering (where text strings are grouped by similarity)
        - Recommendations (where items with related text strings are recommended)
        - Anomaly detection (where outliers with little relatedness are identified)
        - Diversity measurement (where similarity distributions are analyzed)
        - Classification (where text strings are classified by their most similar label)


        Args:
            prompt (str | list[str]): Sentence or list of sentences that will be used for embedding creation.
            encoding_format (Literal[&quot;base64&quot;, &quot;float&quot;], optional): Format of embedding encoding, float - will return in numpay array matrix in other ways it will return bytes representation. Defaults to "float".
            output_type (Optional[Literal[&quot;sentence_embedding&quot;, &quot;token_embeddings&quot;]], optional): The type of embeddings to return: "sentence_embedding" to get sentence embeddings, "token_embeddings" to get wordpiece token embeddings, and None, to get all output values. Defaults to "sentence_embedding".

        Returns:
            EmbeddingResponse: _description_
        """
        if self.verbose:
            panel = Panel(title="|- Embeddings creation request -|", renderable="")
            
            if isinstance(prompt, str):
                panel.renderable += prompt.replace('[', '\[').replace(']', '\]')
            
            else:
                panel.renderable += "\n".join([f"{i+1}. {p.replace('[', '\[').replace(']', '\]')}" for i, p in enumerate(prompt)])
            
            rprint(panel)
        
        async with ClientSession(base_url=self.base_url, headers={"Content-Type": "application/json"}) as client:
            async with client.post("/embeddings", json={
                "sentences": prompt,
                "model": self.model,
                "encoding_format": encoding_format,
                "output_type": output_type,
            }) as resp:
                # For HTTP error handling
                self._exception_manager.listen_for_errors("embeddings.create", resp)

                # NOTE: If http error will be raised, then exception manager will stop execution of code from this line
                response = await resp.json()
                embeddings_data = response["data"]
                usage = response["usage"]

                if encoding_format == "float":
                    embeddings_data = array(embeddings_data)
                else:
                    embeddings_data = embeddings_data.encode()
                
                del response["data"]
                del response["usage"]

                response = EmbeddingResponse(**response, data=embeddings_data, usage=EmbeddingUsage(
                    **usage
                ))
                return response


from langchain_core.callbacks import AsyncCallbackManagerForLLMRun, CallbackManagerForLLMRun
from ascenderai.ascender_ai import AscenderAIChat
from langchain_core.outputs import GenerationChunk, LLMResult, Generation
from langchain_core.language_models.llms import LLM

from typing import Any, AsyncIterator, List, Mapping, Optional
from langchain_core.pydantic_v1 import Field

class AscenderAIL(LLM):
    model_name: str = "mistral-7b-instruct"
    prefix_messages: list[str] = Field(default=[], description="Messages to prefix to the prompt")
    """Series of messages to input before the given message"""
    streaming: bool = Field(default=False, description="Whether to stream the response or not")
    """Whether to stream the response or not"""
    max_tokens: int = Field(default=100, description="Maximum number of tokens to generate")
    """Maximum number of tokens to generate"""

    @property
    def _llm_type(self):
        return "ascender-ai"
    
    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        """Get the identifying parameters."""
        return {
            "model_name": self.model_name,
            "prefix_messages": self.prefix_messages,
            "streaming": self.streaming,
            "max_tokens": self.max_tokens
        }

    async def _astream(self, prompt: str, stop: List[str] | None = None, run_manager: AsyncCallbackManagerForLLMRun | None = None, **kwargs: Any) -> AsyncIterator[GenerationChunk]:
        _client = AscenderAIChat(self.model_name, max_tokens=self.max_tokens, streaming=True, verbose=self.verbose)
        
        response = await _client.create(prompt, stop_seqs=stop, **kwargs)
        async for chunk in response:
            yield GenerationChunk(text=chunk.choices[0].text, generation_info=chunk.model_dump())
            if run_manager:
                await run_manager.on_llm_new_token(chunk.choices[0].text, chunk=chunk.model_dump())

    async def _agenerate(self, prompts: List[str], stop: List[str] | None = None, run_manager: AsyncCallbackManagerForLLMRun | None = None, **kwargs: Any) -> LLMResult:
        _client = AscenderAIChat(self.model_name, max_tokens=self.max_tokens, verbose=self.verbose)
        completions = []
        for prompt in prompts:
            print(self.predict_messages)
            if len(self.prefix_messages):
                prompt = "\n".join(self.prefix_messages) + "\n" + prompt
            completion = await _client.create(prompt, stop_seqs=stop, **kwargs)
            completions.append(list(map(lambda c: Generation(text=c.text, generation_info=c.model_dump()), completion.choices)))
                
        return LLMResult(generations=completions, llm_output={"model_name": self.model_name})

    async def _acall(self, prompt: str, stop: List[str] | None = None, run_manager: AsyncCallbackManagerForLLMRun | None = None, **kwargs: Any) -> str:
        return await self._agenerate([prompt], stop=stop, run_manager=run_manager, **kwargs)
    
    def _call(self, prompt: str, stop: List[str] | None = None, run_manager: CallbackManagerForLLMRun | None = None, **kwargs: Any) -> str:
        raise NotImplementedError("This LLM does not support synchronous calls")


from aiohttp import ClientResponse
from ascenderai.errors import BaseAscenderAIException, ContextWindowExceededError, GatewayTimeoutError, InternalServerError, InvalidRequestFormatError, ServiceUnavailableError


class AscenderExceptionManager:
    def __init__(self, model: str) -> None:
        self.model = model
    
    def listen_for_errors(self, object: str, response: ClientResponse):
        handler = AscenderErrorHandler(object, self.model)
        handler(response)

class AscenderErrorHandler:
    def __init__(self, object: str, model: str) -> None:
        self.object = object
        self.model = model

    def __call__(self, response: ClientResponse) -> None:
        if response.status == 200:
            return
        
        match response.status:
            case 429:
                raise ContextWindowExceededError(self.object, self.model, response.reason)
            case 422:
                raise InvalidRequestFormatError(self.object, self.model, response.reason)
            case 500:
                raise InternalServerError(self.object, self.model)
            case 503:
                raise ServiceUnavailableError(self.object, self.model)
            case 504:
                raise GatewayTimeoutError(self.object, self.model)
            case _:
                raise BaseAscenderAIException(self.object, self.model, f"HTTP {response.status}: {response.reason}")


class BaseAscenderAIException(Exception):
    object: str
    model: str

    def __init__(self, object: str, model: str, *args, **kwargs) -> None:
        self.object = object
        self.model = model
        super().__init__(*args, **kwargs)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(object={self.object!r}, model={self.model!r})"

class ContextWindowExceededError(BaseAscenderAIException):
    def __init__(self, object: str, model: str, message: str) -> None:
        self.message = message
        super().__init__(object, model)

    def __str__(self) -> str:
        return f"HTTP 429: {self.message}"


class InvalidRequestFormatError(BaseAscenderAIException):
    def __init__(self, object: str, model: str, reason: str) -> None:
        self.reason = reason
        super().__init__(object=object, model=model)
    
    def __str__(self) -> str:
        return f"HTTP 422: {self.reason}"


class InternalServerError(BaseAscenderAIException):
    def __str__(self) -> str:
        return "HTTP 500: Internal server error, please try again or contact your AscenderAI Cluster Provider."


class ServiceUnavailableError(BaseAscenderAIException):
    def __str__(self) -> str:
        return "HTTP 503: Services of AscenderAI Cluster aren't available currently, please try again later."
    

class GatewayTimeoutError(BaseAscenderAIException):
    def __str__(self) -> str:
        return "HTTP 504: Gateway Timeout, please try again later."